  integer, parameter :: NTMAX=60
  integer, parameter :: NMAX=NTMAX*12000 !Total sample intervals (one minute)
  integer, parameter :: NDMAX=NTMAX*1500 !Sample intervals at 1500 Hz rate
  integer, parameter :: NSMAX=6827       !Max length of saved spectra
  integer, parameter :: MAXFFT3=16384
