#ifndef N3FJP_H
#define N3FJP_H


// Mapping between ADIF field to N3FJP field
static const QMap<QString, QString> N3FJP_ADIF_MAP = {
    {"AGE", "fldAge"},
    //{"", "fldARCI"},
    {"BAND", "fldBand"},
    {"CALL", "fldCall"},
    //{"", "fldCategory"},
    {"CHECK", "fldCheck"},
    {"CLASS", "fldClass"},
    {"COMMENT", "fldComments"},
    {"COMMENT_INTL", "fldComments"},
    //{"", "fldComputerName"},
    {"CONTEST_ID", "fldContestID"},
    {"CONT", "fldContinent"},
    {"DXCC", "fldCountryDXCC"},
    //{"", "fldCountryWorked"},
    {"COUNTRY", "fldCountyR"},
    {"COUNTRY_INTL", "fldCountyR"},
    {"MY_COUNTRY", "fldCountyS"},
    {"MY_COUNTRY_INTL", "fldCountyS"},
    {"CQZ", "fldCQZone"},
    {"QSO_DATE", "fldDateStr"},
    {"FISTS", "fldFists"},
    {"FISTS_CC", "fldFists"},
    {"FREQ", "fldFrequency"},
    //{"", "fldFuture1"},
    //{"", "fldFuture2"},
    {"GRIDSQUARE", "fldGridR"},
    {"MY_GRIDSQUARE", "fldGridS"},
    //{"", "fldIARUZone"},
    //{"", "fldInitials"},
    {"IOTA", "fldIOTA"},
    {"ITUZ", "fldITUZone"},
    //{"", "fldLightHouse"},
    {"MODE", "fldMode"},
    {"CONTEST_ID", "fldModeContest"},
    {"NAME", "fldNameR"},
    {"NAME_INTL", "fldNameR"},
    {"MY_NAME", "fldNameS"},
    {"MY_NAME_INTL", "fldNameS"},
    {"OPERATOR", "fldOperator"},
    {"*1", "fldOther1"},
    {"*2", "fldOther2"},
    {"*3", "fldOther3"},
    {"*4", "fldOther4"},
    {"*5", "fldOther5"},
    {"*6", "fldOther6"},
    {"*7", "fldOther7"},
    {"*8", "fldOther8"},
    //{"", "fldPoints"},
    {"RX_PWR", "fldPower"},
    {"TX_PWR", "fldPower"},
    {"PRECEDENCE", "fldPrecedence"},
    {"PFX", "fldPrefix"},
    {"PROP_MODE", "fldPropMode"},
    {"QSL_RCVD_VIA", "fldQSLConfByR"},
    {"QSL_SENT_VIA", "fldQSLConfByS"},
    {"QSL_RCVD", "fldQSLR"},
    {"QSL_SENT", "fldQSLS"},
    {"QTH", "fldQTHGroup"},
    {"QTH_INTL", "fldQTHGroup"},
    {"RST_RCVD", "fldRstR"},
    {"RST_SENT", "fldRstS"},
    {"SAT_NAME", "fldSatName"},
    {"ARRL_SECT", "fldSection"},
    {"SRX", "fldSerialNoR"},
    {"STX", "fldSerialNoS"},
    {"SRX_STRING", "fldSPC"},
    {"SRX", "fldSPCNum"},
    {"STATE", "fldState"},
    {"STATION_CALLSIGN", "fldStation"},
    {"TEN_TEN", "fldTenTen"},
    {"TIME_OFF", "fldTimeOffStr"},
    {"TIME_ON", "fldTimeOnStr"},
    //{"", "fldTransmitterID"},
};

#endif // N3FJP_H
